# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
from einops.layers.torch import Rearrange


class DeConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        scale_factor: int = 2,
    ):
        super().__init__()
        self.upscale = nn.UpsamplingNearest2d(scale_factor=scale_factor)
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=kernel_size // 2,
        )
        self.weight = self.conv.weight
        self.bias = self.conv.bias
        self.net = nn.Sequential(self.upscale, self.conv)

    def forward(self, x):
        return self.net(x)


class CNNEncoder(nn.Module):
    # based on https://avandekleut.github.io/vae/
    def __init__(self, ks: int = 3, post_bn: bool = False):
        super(CNNEncoder, self).__init__()

        stride = 2
        padding = ks // 2
        self.add_dim = Rearrange("b h w -> b 1 h w")
        self.add_padding = nn.ZeroPad2d(2)
        self.enc_conv1 = nn.Conv2d(
            1, 2, kernel_size=ks, stride=stride, padding=padding
        )
        nn.init.kaiming_normal_(self.enc_conv1.weight, nonlinearity="relu")
        self.enc_act1 = nn.ReLU()

        self.enc_conv2 = nn.Conv2d(
            2, 4, kernel_size=ks, stride=stride, padding=padding
        )
        nn.init.kaiming_normal_(self.enc_conv2.weight, nonlinearity="relu")
        self.enc_act2 = nn.ReLU()

        if post_bn:
            self.enc_bn1 = nn.BatchNorm2d(num_features=2)
            self.enc_bn2 = nn.BatchNorm2d(num_features=4)
        else:
            self.enc_bn1 = nn.Identity()
            self.enc_bn2 = nn.Identity()

        self.encoder = nn.Sequential(
            self.add_dim,  # 28x28 -> 1x28x28
            self.add_padding,  # 1x28x28 -> 1x32x32
            self.enc_conv1,  # 1x32x32 -> 1x16x16
            self.enc_act1,
            self.enc_bn1,
            self.enc_conv2,  # 1x16x16 -> 1x8x8
            self.enc_act2,
            self.enc_bn2,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.encoder(x)
        return z


class CNNDecoder(nn.Module):
    def __init__(self, ks: int = 3, post_bn: bool = False):
        super(CNNDecoder, self).__init__()

        self.dec_deconv1 = DeConv2d(4, 2, kernel_size=ks, stride=1)
        nn.init.kaiming_normal_(self.dec_deconv1.weight, nonlinearity="relu")
        self.dec_act1 = nn.ReLU()

        self.dec_deconv2 = DeConv2d(2, 1, kernel_size=ks, stride=1)
        nn.init.kaiming_normal_(self.dec_deconv2.weight, nonlinearity="relu")
        # self.dec_act2 = nn.Sigmoid()
        self.rm_padding = nn.ZeroPad2d(-2)
        self.rm_dim = Rearrange("b 1 h w -> b h w")

        if post_bn:
            self.dec_bn1 = nn.BatchNorm2d(2)
            # self.dec_bn2 = nn.BatchNorm2d(1)
        else:
            self.dec_bn1 = nn.Identity()
            # self.dec_bn2 = nn.Identity()

        self.decoder = nn.Sequential(
            self.dec_deconv1,  # 1x8x8 -> 1x16x16
            self.dec_act1,
            self.dec_bn1,
            self.dec_deconv2,  # 1x16x16 -> 1x32x32
            self.rm_padding,  # 1x32x32 -> 1x28x28
            # self.dec_act2,
            # self.dec_bn2,
            self.rm_dim,  # 1x28x28 -> 28x28
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        x_hat = self.decoder(z)
        return x_hat


class DenseEncoder(nn.Module):
    # based on https://avandekleut.github.io/vae/
    def __init__(
        self,
        n_in: int,
        n_hidden1: int,
        n_hidden2: int,
        post_bn: bool = False,
        flatten_input: bool = False,
    ):
        super(DenseEncoder, self).__init__()

        self.enc_dense1 = nn.Linear(n_in, n_hidden1)
        nn.init.kaiming_normal_(self.enc_dense1.weight, nonlinearity="relu")
        self.enc_act1 = nn.ReLU()

        self.enc_dense2 = nn.Linear(n_hidden1, n_hidden2)
        nn.init.kaiming_normal_(self.enc_dense2.weight, nonlinearity="relu")
        self.enc_act2 = nn.ReLU()

        if post_bn:
            self.enc_bn1 = nn.BatchNorm1d(num_features=n_hidden1)
            self.enc_bn2 = nn.BatchNorm1d(num_features=n_hidden2)
        else:
            self.enc_bn1 = nn.Identity()
            self.enc_bn2 = nn.Identity()

        if flatten_input:
            self.rectangle2flat = Rearrange("b h w -> b (h w)")
        else:
            self.rectangle2flat = nn.Identity()

        self.encoder = nn.Sequential(
            self.rectangle2flat,
            self.enc_dense1,
            self.enc_act1,
            self.enc_bn1,
            self.enc_dense2,
            self.enc_act2,
            self.enc_bn2,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.encoder(x)
        return z


class DenseDecoder(nn.Module):
    def __init__(
        self,
        n_latent: int,
        n_hidden1: int,
        n_hidden2: int,
        n_out: int,
        post_bn: bool = False,
        unflatten_output: bool = False,
        w: int = None,
        h: int = None,
    ):
        super(DenseDecoder, self).__init__()
        self.dec_dense1 = nn.Linear(n_latent, n_hidden1)
        nn.init.kaiming_normal_(self.dec_dense1.weight, nonlinearity="relu")
        self.dec_act1 = nn.ReLU()

        self.dec_dense2 = nn.Linear(n_hidden1, n_hidden2)
        nn.init.kaiming_normal_(self.dec_dense2.weight, nonlinearity="relu")
        self.dec_act2 = nn.ReLU()

        self.dec_dense3 = nn.Linear(n_hidden2, n_out)
        # self.dec_act3 = nn.Sigmoid()

        if post_bn:
            self.dec_bn1 = nn.BatchNorm1d(n_hidden1)
            self.dec_bn2 = nn.BatchNorm1d(n_hidden2)
        else:
            self.dec_bn1 = nn.Identity()
            self.dec_bn2 = nn.Identity()

        if unflatten_output:
            if w is None or h is None:
                raise ValueError(
                    f"If unflatten_output is True, w and h must be specified, got {w=}, {h=}."
                )
            if w * h != n_out:
                raise ValueError(
                    f"If unflatten_output is True, w * h must equal n_out, got {w=}, {h=}, {n_out=}."
                )
            self.flat2rectangle = Rearrange("b (h w) -> b h w", h=h, w=w)
        else:
            self.flat2rectangle = nn.Identity()

        self.decoder = nn.Sequential(
            self.dec_dense1,
            self.dec_act1,
            self.dec_bn1,
            self.dec_dense2,
            self.dec_act2,
            self.dec_bn2,
            self.dec_dense3,
            # self.dec_act3,
            self.flat2rectangle,
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        x_hat = self.decoder(z)
        return x_hat


class CNNAutoEncoder(nn.Module):
    def __init__(self, ks: int = 3) -> None:
        super(CNNAutoEncoder, self).__init__()
        self.encoder = CNNEncoder(ks=ks)
        self.decoder = CNNDecoder(ks=ks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        z = self.encoder(x)
        x_hat = self.decoder(z)
        return x_hat



# -*- coding: utf-8 -*-
import logging
import random

import numpy as np
import torch
from rich.logging import RichHandler


def get_logger(name: str = "rich", level=logging.INFO):
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    logging.basicConfig(
        level=level,
        format="%(name)s: %(levelname)s - %(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


logger = get_logger("utils.py")


def make_deterministic(seed: int = 42):
    # from https://krokotsch.eu/posts/deep-learning-unit-tests/
    # PyTorch
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    # Numpy
    np.random.seed(seed)

    # Built-in Python
    random.seed(seed)



# -*- coding: utf-8 -*-
import typing as T

import numpy as np
import torch
import tqdm
from sklearn import base


def sigmoid(z: torch.Tensor) -> torch.Tensor:
    return 1 / (1 + torch.exp(-z))


def sigmoid_derivative(z: torch.Tensor) -> torch.Tensor:
    exp = torch.exp(z)
    return exp / (1 + exp) ** 2


class Rumelhart1986PerceptronClassifier(
    base.BaseEstimator, base.ClassifierMixin
):
    def __init__(
        self,
        n_hidden: T.Tuple[int] = (10, 5),
        eps: float = 0.1,
        alpha: float = 0.0,
        random_state: int = 42,
        verbose: bool = False,
        epochs: int = 10,
    ) -> None:
        self.n_hidden = n_hidden
        self.eps = eps
        self.random_state = random_state
        self.verbose = verbose
        self.epochs = epochs
        self.alpha = alpha

    def _handle_Xy(
        self, X: np.ndarray, y: np.ndarray = None
    ) -> T.Tuple[torch.tensor, T.Optional[torch.tensor]]:
        _X = torch.from_numpy(X).double()  # (N_samples, N_features)
        if y is None:
            return _X, None
        _y = torch.from_numpy(y).double()  # (N_samples,)
        return _X, _y

    @torch.no_grad()
    def _forward(self, X: torch.tensor, training: bool = False) -> torch.tensor:
        a = torch.concat(
            (X, torch.ones((X.shape[0], 1), dtype=torch.double)), dim=1
        )

        if training:
            self.telemetry_["z"] = []
            self.telemetry_["a"] = [a]

        for w in self.weights_:
            z = a @ w
            a = sigmoid(z)

            if training:
                self.telemetry_["z"].append(z)
                self.telemetry_["a"].append(a)

        return a

    @torch.no_grad()
    def _backward(self, y_pred: torch.tensor, y: torch.tensor) -> None:
        # MSE loss derivative
        delta = y_pred - y.view((-1, 1))

        # iterating through the layers in reverse order
        # from the output layer to the input layer
        ix_layers = list(range(len(self.weights_)))[::-1]
        gradients = [None for _ in ix_layers]
        if self.alpha > 0:
            old_gradients = (
                self.telemetry_["gradients"]
                if "gradients" in self.telemetry_
                else None
            )

        for i in ix_layers:
            z = self.telemetry_["z"][i]
            a = self.telemetry_["a"][i]

            delta = delta * sigmoid_derivative(z)

            # part to be added to the weight for layer i
            dw = a.T @ delta
            gradients[i] = dw

            w = self.weights_[i]
            delta = delta @ w.T

        for i in ix_layers:
            dw = self.eps * gradients[i]

            if self.alpha > 0 and old_gradients is not None:
                dw += self.alpha * old_gradients[i]

            self.weights_[i] -= dw

        if self.alpha > 0:
            self.telemetry_["gradients"] = gradients

    @torch.no_grad()
    def fit(
        self, X: np.ndarray, y: np.ndarray
    ) -> "Rumelhart1986PerceptronClassifier":
        torch.manual_seed(self.random_state)

        self.units_ = (
            [X.shape[1] + 1] + [h + 1 for h in self.n_hidden] + [1]
        )  # +1 because of bias
        self.weights_ = [
            torch.randn(self.units_[i], self.units_[i + 1], dtype=torch.double)
            for i in range(len(self.units_) - 1)
        ]
        self.telemetry_ = dict()
        self.errors_ = torch.zeros(
            self.epochs, dtype=torch.float32
        )  # (N_epochs,)
        X, y = self._handle_Xy(X, y)

        for epoch in tqdm.tqdm(range(self.epochs), disable=not self.verbose):
            y_pred = self._forward(X, training=True)
            self._backward(y_pred, y)
            self.errors_[epoch] = torch.mean((y_pred - y) ** 2)

        return self

    @torch.no_grad()
    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        X, _ = self._handle_Xy(X)
        y_pred = self._forward(X, training=False)
        return y_pred.detach().numpy()

    @torch.no_grad()
    def predict(self, X: np.ndarray) -> np.ndarray:
        y_pred = self.predict_proba(X)
        return y_pred > 0.5



# -*- coding: utf-8 -*-
import typing as T

import numpy as np
import torch
import torch.nn.functional as F
import tqdm
from scipy.sparse import issparse
from sklearn import base, preprocessing
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted

import random_neural_net_models.utils as utils

logger = utils.get_logger("perceptron", level="DEBUG")


class PerceptronClassifier(base.BaseEstimator, base.ClassifierMixin):
    """OG neural net

    Implementation follows: https://en.wikipedia.org/wiki/Perceptron
    """

    def __init__(
        self,
        learning_rate: float = 0.01,
        epochs: int = 1000,
        random_state: int = 42,
        verbose: bool = False,
    ):
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.random_state = random_state
        self.verbose = verbose

    def _handle_Xy(
        self, X: np.ndarray, y: np.ndarray = None
    ) -> T.Tuple[torch.tensor, T.Optional[torch.tensor]]:
        _X = torch.from_numpy(X).double()  # (N_samples, N_features)
        if y is None:
            return _X, None
        _y = torch.from_numpy(y).double()  # (N_samples,)
        return _X, _y

    @torch.no_grad()
    def _fit_two_class(self, X: np.ndarray, y: np.ndarray):
        torch.manual_seed(self.random_state)

        self.weights_ = torch.randn(
            X.shape[1], dtype=torch.double
        )  # (N_features,)
        self.bias_ = torch.zeros(1, dtype=torch.double)  # (1,)

        _X, _y = self._handle_Xy(X, y)
        self.errors_ = torch.zeros(
            self.epochs, dtype=torch.float32
        )  # (N_epochs,)

        for epoch in tqdm.tqdm(range(self.epochs), disable=not self.verbose):
            y_hat = _X @ self.weights_ + self.bias_

            y_hat_discrete = torch.where(
                y_hat > 0, 1.0, 0.0
            )  # heaviside step function

            dy = _y - y_hat_discrete
            dw = self.learning_rate * _X.T @ dy
            y_error = torch.sum(dy)
            db = self.learning_rate * y_error

            self.weights_ += dw
            self.bias_ += db

            self.errors_[epoch] = y_error.abs()

    @torch.no_grad()
    def _fit_multi_class(self, X: np.ndarray, y: np.ndarray):
        torch.manual_seed(self.random_state)

        # different to two-class perceptron, we need to have a weight vector for each class
        self.weights_ = torch.randn(
            (X.shape[1], self.n_classes_), dtype=torch.double
        )  # (N_features,N_classes)
        self.bias_ = torch.zeros(1, dtype=torch.double)  # (1,)

        _X, _y = self._handle_Xy(X, y)
        _y_int = _y.clone().long()
        # we convert the integer labels to one-hot vectors to be able to apply the same math as for the two-class perceptron
        y_onehot = F.one_hot(_y_int, num_classes=self.n_classes_)
        self.errors_ = torch.zeros(
            self.epochs, dtype=torch.float32
        )  # (N_epochs,)

        for epoch in tqdm.tqdm(range(self.epochs), disable=not self.verbose):
            y_hat = _X @ self.weights_ + self.bias_

            # different to two-class perceptron, we select the class with the highest score
            # directly, without using the heaviside step function
            y_hat_discrete = torch.argmax(y_hat, dim=1)

            # now onehot encoding the discrete predictions
            y_hat_onehot = F.one_hot(
                y_hat_discrete, num_classes=self.n_classes_
            )

            dy = y_onehot - y_hat_onehot

            dw = self.learning_rate * _X.T @ dy.double()

            y_error = torch.sum(
                dy.abs()
            )  # TODO: why no convergence if .abs is missing here?
            db = self.learning_rate * y_error

            self.weights_ += dw
            self.bias_ += db

            self.errors_[epoch] = y_error

    def fit(self, X: np.ndarray, y: np.ndarray) -> "PerceptronClassifier":
        if issparse(y):
            raise ValueError(
                "sparse multilabel-indicator for y is not supported."
            )

        X, y = self._validate_data(
            X,
            y,
            ensure_2d=True,
            y_numeric=base.is_regressor(self),
            multi_output=False,
            dtype="numeric",
            allow_nd=True,
        )

        self.n_features_in_ = X.shape[1]

        self.classes_ = unique_labels(y)  # np.unique(y)
        self.n_classes_ = len(self.classes_)
        self.is_multi_class_ = self.n_classes_ > 2
        self.y_enc_ = preprocessing.LabelEncoder().fit(y)
        y = self.y_enc_.transform(y)

        if self.is_multi_class_:
            logger.debug(
                f"More than two classes detected ({self.classes_}), treating as multi-class problem."
            )
            self._fit_multi_class(X, y)
        else:
            logger.debug(
                f"Two classes detected ({self.classes_}), treating as two-class problem."
            )
            self._fit_two_class(X, y)

        return self

    @torch.no_grad()
    def predict(self, X: np.ndarray) -> np.ndarray:
        check_is_fitted(
            self,
            (
                "classes_",
                "n_classes_",
                "is_multi_class_",
                "weights_",
                "bias_",
                "n_features_in_",
                "errors_",
            ),
        )
        X = self._validate_data(X, dtype=[np.float32, np.float64], reset=False)

        if X.shape[1] != self.n_features_in_:
            raise ValueError(f"{X.shape[1]=} != {self.n_features_in_=}")

        _X, _ = self._handle_Xy(X, y=None)
        y_hat = _X @ self.weights_ + self.bias_

        if self.is_multi_class_:
            y_pred = torch.argmax(y_hat, dim=1).detach().numpy()

        else:
            y_pred = torch.where(y_hat > 0, 1, 0).detach().numpy()

        y_pred = self.y_enc_.inverse_transform(y_pred)
        return y_pred



# -*- coding: utf-8 -*-
# based on https://github.com/fastai/course22p2/blob/master/nbs/26_diffusion_unet.ipynb
import typing as T

import torch
import torch.nn as nn
from einops.layers.torch import Rearrange

import random_neural_net_models.utils as utils

logger = utils.get_logger("resnet.py")


def get_conv_pieces(
    num_features_in: int, num_features_out: int, kernel_size: int, stride: int
) -> T.Tuple[nn.Module, nn.Module, nn.Module]:
    """Batch norm, SiLU activation and conv2d layer for the unet's resnet blocks."""
    bn = nn.BatchNorm2d(num_features=num_features_in)
    act = nn.SiLU()
    padding = kernel_size // 2
    conv = nn.Conv2d(
        num_features_in,
        num_features_out,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
    )
    return bn, act, conv


class ResBlock(nn.Module):
    def __init__(
        self,
        num_features_in: int,
        num_features_out: int,
        stride: int = 1,
        ks: int = 3,
    ):
        super().__init__()

        self.bn1, self.act1, self.conv1 = get_conv_pieces(
            num_features_in, num_features_out, ks, stride=1
        )
        self.bn2, self.act2, self.conv2 = get_conv_pieces(
            num_features_out, num_features_out, ks, stride=stride
        )

        self.convs = nn.Sequential(
            self.bn1,
            self.act1,
            self.conv1,
            self.bn2,
            self.act2,
            self.conv2,
        )

        self.use_identity = num_features_in == num_features_out
        self.idconv = (
            nn.Identity()
            if self.use_identity
            else nn.Conv2d(
                num_features_in, num_features_out, kernel_size=1, stride=1
            )
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_conv = self.convs(x)
        x_id = self.idconv(x)
        return x_conv + x_id


class SaveModule:
    def forward(self, x, *args, **kwargs):
        self.saved_output = super().forward(x, *args, **kwargs)
        return self.saved_output


class SavedResBlock(SaveModule, ResBlock):
    pass


class DownBlock(nn.Module):
    def __init__(
        self,
        num_features_in: int,
        num_features_out: int,
        add_down: bool = True,
        num_resnet_layers: int = 1,
    ):
        """Sequence of resnet blocks with a downsample at the end, see stride."""
        super().__init__()

        self.add_down = add_down
        self.setup_res_blocks(
            num_features_in,
            num_features_out,
            num_resnet_layers=num_resnet_layers,
        )

        self.setup_downscaling(num_features_out)

    def setup_res_blocks(
        self,
        num_features_in: int,
        num_features_out: int,
        num_resnet_layers: int = 2,
    ):
        self.res_blocks = nn.ModuleList()
        for i in range(num_resnet_layers - 1):
            n_in = num_features_in if i == 0 else num_features_out
            self.res_blocks.append(ResBlock(n_in, num_features_out))

        self.res_blocks.append(
            SavedResBlock(
                num_features_in=num_features_out,
                num_features_out=num_features_out,
            )
        )

    def setup_downscaling(self, num_features_out: int):
        if self.add_down:
            self.down = nn.Conv2d(
                num_features_out, num_features_out, 3, stride=2, padding=1
            )
        else:
            self.down = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for res_block in self.res_blocks:
            x = res_block(x)
        return self.down(x)

    @property
    def saved_output(self):
        return self.res_blocks[-1].saved_output


class UNetDown(nn.Module):
    def __init__(self, num_features: T.Tuple[int], num_layers: int) -> None:
        super().__init__()

        n_ins = [num_features[0]] + list(num_features[:-1])
        n_outs = [num_features[0]] + list(num_features[1:])
        add_downs = [True] * (len(num_features) - 1) + [False]

        self.down_blocks = nn.ModuleList(
            [
                DownBlock(
                    n_in, n_out, add_down=add_down, num_resnet_layers=num_layers
                )
                for n_in, n_out, add_down in zip(n_ins, n_outs, add_downs)
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for down_block in self.down_blocks:
            x = down_block(x)
        return x

    def __iter__(self) -> torch.Tensor:
        for down_block in self.down_blocks:
            yield down_block.saved_output


class UpBlock(nn.Module):
    def __init__(
        self,
        num_features_in: int,
        prev_num_features_out: int,
        num_features_out: int,
        add_up: bool = True,
        num_resnet_layers: int = 2,
    ):
        super().__init__()
        self.add_up = add_up
        self.setup_res_blocks(
            num_features_in,
            prev_num_features_out,
            num_features_out,
            num_resnet_layers=num_resnet_layers,
        )

        self.setup_upscaling(num_features_out)

    def setup_res_blocks(
        self,
        num_features_in: int,
        prev_num_features_out: int,
        num_output_features: int,
        num_resnet_layers: int = 2,
    ):
        self.res_blocks = nn.ModuleList()
        n_out = num_output_features

        for i in range(num_resnet_layers):
            if i == 0:
                n_in = prev_num_features_out
            else:
                n_in = n_out

            # handling unet copy
            if i == 0:
                n_in += num_features_in

            self.res_blocks.append(ResBlock(n_in, n_out))

    def setup_upscaling(self, num_features_out: int):
        if self.add_up:
            self.up = nn.Sequential(
                nn.Upsample(scale_factor=2, mode="nearest"),
                nn.Conv2d(
                    num_features_out, num_features_out, kernel_size=3, padding=1
                ),
            )
        else:
            self.up = nn.Identity()

    def forward(
        self, x_up: torch.Tensor, xs_down: T.List[torch.Tensor]
    ) -> torch.Tensor:
        x_glue = torch.cat([x_up, xs_down.pop()], dim=1)
        x = self.res_blocks[0](x_glue)

        for res_block in self.res_blocks[1:]:
            x = res_block(x)

        if self.add_up:
            return self.up(x)
        else:
            return x


class UNetUp(nn.Module):
    def __init__(
        self,
        downs: UNetDown,
    ) -> None:
        super().__init__()

        self.ups = nn.ModuleList()
        n = len(downs.down_blocks)
        up_block = None

        for i, down_block in enumerate(reversed(downs.down_blocks)):
            is_final_layer = i == n - 1

            # 3 infos we need:
            # n_in_down: input features from parallel down block
            # n_in_prev_up: input features from previous up block
            # n_out_up: output features of current up block

            # n_in_down
            if not is_final_layer:  # res block
                down_out_conv = down_block.res_blocks[-1].convs[2]
            else:  # down conv
                down_out_conv = down_block.down

            n_in_down = down_out_conv.out_channels

            # n_in_prev_up
            if up_block is None:
                n_in_prev_up = n_in_down
            elif up_block.add_up:  # up conv
                n_in_prev_up = up_block.up[1].out_channels
            else:
                raise ValueError(f"unexpected case for {up_block=}")

            # n_out_up
            down_input_conv = down_block.res_blocks[0].convs[
                2
            ]  # (bn, act, conv)
            n_out_up = down_input_conv.in_channels

            add_up = not is_final_layer

            num_resnet_layers = len(down_block.res_blocks)

            up_block = UpBlock(
                n_in_down,
                n_in_prev_up,
                n_out_up,
                add_up=add_up,
                num_resnet_layers=num_resnet_layers,
            )
            self.ups.append(up_block)

    def forward(
        self, x: torch.Tensor, saved: T.List[torch.Tensor]
    ) -> torch.Tensor:
        for upblock in self.ups:
            x = upblock(x, saved)
        return x


class UNetModel(nn.Module):
    def __init__(
        self,
        in_channels: int = 1,
        out_channels: int = 1,
        list_num_features: T.Tuple[int] = (8, 16),
        num_layers: int = 2,
    ):
        super().__init__()
        if in_channels != out_channels:
            logger.warning(
                f"in_channels ({in_channels}) != out_channels ({out_channels})"
            )

        self.setup_input(in_channels, list_num_features)

        self.downs = UNetDown(list_num_features, num_layers=num_layers)

        self.mid_block = ResBlock(list_num_features[-1], list_num_features[-1])

        self.ups = UNetUp(self.downs)

        self.setup_output(list_num_features, out_channels)

    def setup_input(self, in_channels: int, list_num_features: T.Tuple[int]):
        if in_channels == 1:
            self.add_dim = Rearrange("b h w -> b 1 h w")
        else:
            self.add_dim = nn.Identity()

        self.add_padding = nn.ZeroPad2d(2)
        self.conv_in = nn.Conv2d(
            in_channels, list_num_features[0], kernel_size=3, padding=1
        )
        self.wrangle_input = nn.Sequential(
            self.add_dim, self.add_padding, self.conv_in
        )

    def setup_output(self, list_num_features: T.Tuple[int], out_channels: int):
        self.bn_out, self.act_out, self.conv_out = get_conv_pieces(
            list_num_features[0], out_channels, kernel_size=1, stride=1
        )

        if out_channels == 1:
            self.rm_dim = Rearrange("b 1 h w -> b h w")
        else:
            self.rm_dim = nn.Identity()

        self.rm_padding = nn.ZeroPad2d(-2)

        self.wrangle_output = nn.Sequential(
            self.bn_out,
            self.act_out,
            self.conv_out,
            self.rm_dim,
            self.rm_padding,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # input
        x = self.wrangle_input(x)
        saved = [x]

        # down projections
        x = self.downs(x)

        # copy from down projections for up projections
        saved.extend([output for output in self.downs])

        x = self.mid_block(x)

        # up projections
        x = self.ups(x, saved)

        # output
        x = self.wrangle_output(x)

        return x



# -*- coding: utf-8 -*-
import re
import typing as T
from dataclasses import dataclass

import torch.nn as nn

import random_neural_net_models.utils as utils

logger = utils.get_logger("search.py")


@dataclass
class NamedModule:
    name: str
    module: nn.Module


def collate_named_modules(
    module: nn.Module,
    module_name: str = None,
    depth: int = 0,
    max_depth: int = 3,
    collated_named_modules: T.List[NamedModule] = None,
) -> T.List[NamedModule]:
    if depth >= max_depth:
        return collated_named_modules
    elif len(list(module.children())) == 0:
        return collated_named_modules

    if collated_named_modules is None:
        collated_named_modules = []

    depth += 1

    for child_name, child_module in module.named_children():
        child_name = (
            f"{module_name}.{child_name}"
            if module_name is not None
            else child_name
        )
        collated_named_modules.append(NamedModule(child_name, child_module))

        collated_named_modules = collate_named_modules(
            child_module,
            module_name=child_name,
            depth=depth,
            max_depth=max_depth,
            collated_named_modules=collated_named_modules,
        )

    return collated_named_modules


def find_named_module(
    collated_named_modules: T.List[NamedModule], pattern: str
) -> T.List[NamedModule]:
    matches = []
    for named_module in collated_named_modules:
        if re.match(pattern, named_module.name) is not None:
            matches.append(named_module)
    if len(matches) > 0:
        return matches
    else:
        logger.error(f"Module with {pattern=} not found")
        return []


class ChildSearch:
    def __init__(self, module: nn.Module, max_depth: int = 3):
        self.module = module
        self.collated_named_modules = collate_named_modules(
            module, max_depth=max_depth
        )

    def __call__(self, *patterns: T.Tuple[str]) -> T.List[NamedModule]:
        matches = []
        for pattern in patterns:
            matches.extend(
                find_named_module(self.collated_named_modules, pattern)
            )
        return matches

    @property
    def names(self) -> T.List[str]:
        return [
            named_module.name for named_module in self.collated_named_modules
        ]



# -*- coding: utf-8 -*-
import typing as T

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops.layers.torch import Rearrange

import random_neural_net_models.autoencoder_fastai2022 as ae


class Variational(nn.Module):
    def __init__(
        self,
        n_in: int,
        n_latent: int,
        flatten_input: bool = True,
        unflatten_output: bool = True,
        c: int = None,
        w: int = None,
        h: int = None,
        post_bn: bool = True,
    ) -> None:
        super(Variational, self).__init__()

        self.mu = nn.Linear(n_in, n_latent)
        self.logvar = nn.Linear(n_in, n_latent)

        if post_bn:
            self.mu_bn_post = nn.BatchNorm1d(n_latent)
            self.logvar_bn_post = nn.BatchNorm1d(n_latent)
        else:
            self.mu_bn_post = nn.Identity()
            self.logvar_bn_post = nn.Identity()

        if flatten_input:
            self.rectangle2flat = Rearrange("b c h w -> b (c h w)")
        else:
            self.rectangle2flat = nn.Identity()

        if unflatten_output:
            if c is None or w is None or h is None:
                raise ValueError(
                    f"If unflatten_output is True, c, w, h must be specified, got {c=}, {w=}, {h=}."
                )
            self.flat2rectangle = Rearrange(
                "b (c h w) -> b c h w", c=c, h=h, w=w
            )
        else:
            self.flat2rectangle = nn.Identity()

    def forward(
        self, x: torch.Tensor
    ) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x = self.rectangle2flat(x)

        mu = self.mu(x)
        logvar = self.logvar(x)
        mu = self.mu_bn_post(mu)
        logvar = self.logvar_bn_post(logvar)

        std = (0.5 * logvar).exp()
        eps = torch.randn_like(std)

        z = mu + eps * std

        z = self.flat2rectangle(z)

        return z, mu, logvar


class CNNVariationalAutoEncoder(nn.Module):
    def __init__(
        self,
        ks: int = 3,
        n_latent: int = 4 * 8 * 8,
        post_bn_encoder: bool = True,
        post_bn_variational: bool = True,
        post_bn_decoder: bool = True,
    ) -> None:
        super(CNNVariationalAutoEncoder, self).__init__()
        self.encoder = ae.CNNEncoder(ks=ks, post_bn=post_bn_encoder)
        n_enc = 4 * 8 * 8
        self.variational = Variational(
            n_in=n_enc,
            n_latent=n_latent,
            flatten_input=True,
            unflatten_output=True,
            c=4,
            w=8,
            h=8,
            post_bn=post_bn_variational,
        )
        self.decoder = ae.CNNDecoder(ks=ks, post_bn=post_bn_decoder)

    def forward(
        self, x: torch.Tensor
    ) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x = self.encoder(x)
        z, mu, logvar = self.variational(x)
        x_hat = self.decoder(z)
        return x_hat, mu, logvar


class DenseVariationalAutoEncoder(nn.Module):
    def __init__(
        self,
        w: int,
        h: int,
        n_latent: int = 200,
        post_bn_encoder: bool = True,
        post_bn_variational: bool = True,
        post_bn_decoder: bool = True,
    ) -> None:
        super(DenseVariationalAutoEncoder, self).__init__()
        n_in = w * h
        n_enc_hidden1 = 400
        n_enc_hidden2 = 200
        self.encoder = ae.DenseEncoder(
            n_in=n_in,
            n_hidden1=n_enc_hidden1,
            n_hidden2=n_enc_hidden2,
            post_bn=post_bn_encoder,
            flatten_input=True,
        )

        self.variational = Variational(
            n_in=n_enc_hidden2,
            n_latent=n_latent,
            flatten_input=False,
            unflatten_output=False,
            post_bn=post_bn_variational,
        )

        n_dec_hidden1 = 200
        n_dec_hidden2 = 400

        self.decoder = ae.DenseDecoder(
            n_latent=n_latent,
            n_hidden1=n_dec_hidden1,
            n_hidden2=n_dec_hidden2,
            n_out=n_in,
            post_bn=post_bn_decoder,
            unflatten_output=True,
            w=w,
            h=h,
        )

    def forward(
        self, x: torch.Tensor
    ) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x = self.encoder(x)
        z, mu, logvar = self.variational(x)
        x_hat = self.decoder(z)
        return x_hat, mu, logvar


class CNNDenseVariationalAutoEncoder(nn.Module):
    def __init__(
        self,
        w: int,
        h: int,
        ks: int = 3,
        n_latent: int = 200,
        post_bn_encoder: bool = True,
        post_bn_variational: bool = True,
        post_bn_decoder: bool = True,
    ) -> None:
        super(CNNDenseVariationalAutoEncoder, self).__init__()
        n_in = w * h

        self.encoder = ae.CNNEncoder(ks=ks, post_bn=post_bn_encoder)

        n_enc = 4 * 8 * 8
        self.variational = Variational(
            n_in=n_enc,
            n_latent=n_latent,
            flatten_input=True,
            unflatten_output=False,
            c=4,
            w=8,
            h=8,
            post_bn=post_bn_variational,
        )

        n_dec_hidden1 = 200
        n_dec_hidden2 = 400

        self.decoder = ae.DenseDecoder(
            n_latent=n_latent,
            n_hidden1=n_dec_hidden1,
            n_hidden2=n_dec_hidden2,
            n_out=n_in,
            post_bn=post_bn_decoder,
            unflatten_output=True,
            w=w,
            h=h,
        )

    def forward(
        self, x: torch.Tensor
    ) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x = self.encoder(x)
        z, mu, logvar = self.variational(x)
        x_hat = self.decoder(z)
        return x_hat, mu, logvar


def calc_distribution_divergence_loss(
    input: T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor], x: torch.Tensor
) -> torch.Tensor:
    _, mu, logvar = input
    s = 1 + logvar - mu.pow(2) - logvar.exp()
    return -0.5 * s.mean()


def calc_reconstruction_loss(
    input: T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    x: torch.Tensor,
    is_sigmoid: bool = False,
) -> torch.Tensor:
    x_hat, _, _ = input
    if is_sigmoid:
        return F.mse_loss(x, x_hat)
    else:
        return F.binary_cross_entropy_with_logits(x_hat, x)


def calc_vae_loss(
    input: T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    x: torch.Tensor,
    is_sigmoid: bool = False,
) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    reconstruction_loss = calc_reconstruction_loss(
        input, x, is_sigmoid=is_sigmoid
    )
    divergence_loss = calc_distribution_divergence_loss(input, x)
    total_loss = reconstruction_loss + divergence_loss
    return total_loss, reconstruction_loss, divergence_loss


def calc_vae_test_loss(
    model_output: T.List[T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]],
    x: torch.Tensor,
    is_sigmoid: bool = False,
) -> T.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    x_hat = torch.cat([_x[0] for _x in model_output], dim=0)
    mu = torch.cat([_x[1] for _x in model_output], dim=0)
    logvar = torch.cat([_x[2] for _x in model_output], dim=0)
    _model_output = (x_hat, mu, logvar)
    reconstruction_loss = calc_reconstruction_loss(
        _model_output, x, is_sigmoid=is_sigmoid
    )
    divergence_loss = calc_distribution_divergence_loss(_model_output, x)
    total_loss = reconstruction_loss + divergence_loss
    return total_loss, reconstruction_loss, divergence_loss


import re


def check_module_name_is_activation(name: str) -> bool:
    return re.match(r".*act\d$", name) is not None


def check_module_name_grad_relevant(name: str) -> bool:
    is_enc_relevant = re.match(r"^enc_(bn|conv|dense)", name) is not None
    is_var_relevant = re.match(r"^mu|logvar|bn_post", name) is not None
    is_dec_relevant = re.match(r"^dec_(bn|deconv|dense)", name) is not None
    is_not_act = re.match(r".*act\d$", name) is None
    return (
        is_enc_relevant or is_var_relevant or is_dec_relevant
    ) and is_not_act



# -*- coding: utf-8 -*-
import typing as T
from collections import defaultdict
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn

import random_neural_net_models.search as search
import random_neural_net_models.utils as utils

logger = utils.get_logger("telemetry.py")


class History:
    def __init__(
        self,
        model: nn.Module,
        every_n: int = 1,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        self.not_initialized = name_patterns is None or len(name_patterns) == 0
        if self.not_initialized:
            logger.info("Not collecting history.")
            return

        self.child_search = search.ChildSearch(
            model, max_depth=max_depth_search
        )
        self.every_n = every_n
        self.modules = self.child_search(*name_patterns)

    @property
    def name_matches(self):
        if self.not_initialized:
            return []
        return [a.name for a in self.modules]

    def clean(self):
        if self.not_initialized:
            logger.info("No activation hooks to clean up.")
            return
        for hook in self.hooks.values():
            hook.remove()
        self.hooks.clear()


class ActivationsHistory(History):
    def __init__(
        self,
        model: nn.Module,
        every_n: int = 1,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        super().__init__(model, every_n, name_patterns, max_depth_search)

        self.stats: T.Dict[str, T.List[ActivationStats]] = defaultdict(list)
        self.register_hooks(model, name_patterns, max_depth_search)

    def register_hooks(
        self,
        model: nn.Module,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        self.hooks = defaultdict(None)

        for named_module in self.modules:
            cas = CollectorActivationStats(
                self, named_module.name, every_n=self.every_n
            )
            self.hooks[
                named_module.name
            ] = named_module.module.register_forward_hook(cas)

        logger.info(
            f"Will collect activiation history every {self.every_n}th iteration for: {self.name_matches=}"
        )


class GradientsHistory(History):
    def __init__(
        self,
        model: nn.Module,
        every_n: int = 1,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        super().__init__(model, every_n, name_patterns, max_depth_search)
        self.stats: T.Dict[str, T.List[ParameterStats]] = defaultdict(list)
        self.register_hooks(model, name_patterns, max_depth_search)

    def register_hooks(
        self,
        model: nn.Module,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        self.child_search = search.ChildSearch(
            model, max_depth=max_depth_search
        )
        self.modules = self.child_search(*name_patterns)
        self.hooks = defaultdict(None)

        for named_module in self.modules:
            cgs = CollectorGradientStats(
                self, named_module.name, every_n=self.every_n
            )
            self.hooks[
                named_module.name
            ] = named_module.module.register_full_backward_hook(cgs)

        logger.info(
            f"Will collect gradient history every {self.every_n}th iteration for: {self.name_matches=}"
        )


class HistoryException(Exception):
    pass


class ParametersHistory(History):
    def __init__(
        self,
        model: nn.Module,
        every_n: int = 1,
        name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        super().__init__(model, every_n, name_patterns, max_depth_search)
        self.stats: T.Dict[str, T.List[ParameterStats]] = defaultdict(list)
        self.every_n = every_n
        logger.info(
            f"Will collect parameter history every {self.every_n}th iteration for: {self.name_matches=}"
        )

    def __call__(self, _iter: int):
        if self.not_initialized or _iter % self.every_n != 0:
            return

        for named_module in self.modules:
            state_dict = named_module.module.state_dict()
            for name, parameter_values in state_dict.items():
                parameter_values = parameter_values.detach().flatten().float()
                mean = parameter_values.mean().cpu().item()
                std = parameter_values.std().cpu().item()
                abs_perc90 = parameter_values.abs().quantile(0.9).cpu().item()
                self.stats[f"{named_module.name}.{name}"].append(
                    ParameterStats(_iter, mean, std, abs_perc90)
                )

    def draw(
        self,
        name: str,
        figsize: T.Tuple[int, int] = (12, 7),
    ) -> None:
        if self.not_initialized:
            logger.info("Not drawing parameter history.")
            return
        fig, axs = plt.subplots(figsize=figsize, nrows=2, sharex=True)

        ax = axs[0]
        _name = f"{name}.weight"
        df = self.get_df(_name)

        ax.fill_between(
            df["iter"],
            df["mean"] - df["std"],
            df["mean"] + df["std"],
            alpha=0.5,
            label="mean+-std",
        )
        sns.lineplot(
            data=df,
            x="iter",
            y="mean",
            ax=ax,
            label="mean",
        )
        sns.lineplot(
            data=df,
            x="iter",
            y="abs_perc90",
            ax=ax,
            label="90%(abs(param))",
        )
        ax.legend()
        ax.set_ylabel("weight")
        ax.set_title(f"{_name}")

        ax = axs[1]
        _name = f"{name}.bias"
        df = self.get_df(_name)

        is_useful_std = (~(df["std"].isin([np.inf, -np.inf]).any())) & df[
            "std"
        ].notna().all()
        if is_useful_std:
            ax.fill_between(
                df["iter"],
                df["mean"] - df["std"],
                df["mean"] + df["std"],
                alpha=0.5,
                label="mean+-std",
            )
        else:
            logger.error(
                f"{_name=} df['std'] has inf or nan values, skipping fillbetween plot."
            )

        sns.lineplot(
            data=df,
            x="iter",
            y="mean",
            ax=ax,
            label="mean",
        )
        sns.lineplot(
            data=df,
            x="iter",
            y="abs_perc90",
            ax=ax,
            label="90%(abs(param))",
        )
        ax.legend()
        ax.set_ylabel("bias")
        ax.set_title(f"{_name}")

        plt.tight_layout()
        plt.show()

    def get_df(self, name: str) -> pd.DataFrame:
        if self.not_initialized:
            logger.info("Not getting parameter history.")
            return
        df = pd.DataFrame(self.stats[name])

        # print error to log if any column has inf or nan values
        isna = df.isna()
        if df.isna().any().any():
            mean_na = (
                isna.mean().sort_values(ascending=False).rename("fraction")
            )
            mean_na.index.name = "column"
            raise HistoryException(
                f"{name=} df has missing values: {mean_na.to_markdown()}"
            )

        isinf = df.isin([np.inf, -np.inf])
        if isinf.any().any():
            mean_inf = (
                isinf.mean().sort_values(ascending=False).rename("fraction")
            )
            mean_inf.index.name = "column"
            raise HistoryException(
                f"{name=} df has inf values: {mean_inf.to_markdown()}"
            )

        # print error if df is empty
        if len(df) == 0:
            raise HistoryException(f"{name=} df is empty")

        return df


class CollectorActivationStats:
    def __init__(
        self,
        hook: ActivationsHistory,
        name: str,
        every_n: int = 1,
        threshold_dead=1e-6,
    ):
        self.hook = hook
        self.name = name
        self.every_n = every_n
        self.iter = 0
        self.threshold_dead = threshold_dead

    def __call__(
        self,
        module: nn.Module,
        input: torch.Tensor,
        output: torch.Tensor,
    ):
        self.iter += 1
        if self.iter % self.every_n != 0:
            return

        acts = output.detach().flatten()
        mean = acts.mean().cpu().item()
        std = acts.std().cpu().item()
        frac_dead = (acts.abs() < self.threshold_dead).sum().cpu().item() / len(
            acts
        )

        self.hook.stats[self.name].append(ActivationStats(mean, std, frac_dead))


class CollectorGradientStats:
    def __init__(
        self,
        hook: GradientsHistory,
        name: str,
        every_n: int = 1,
        threshold_dead: float = 1e-8,
    ):
        self.hook = hook
        self.name = name
        self.every_n = every_n
        self.iter = 0
        self.threshold_dead = threshold_dead

    def __call__(
        self,
        module: nn.Module,
        input: torch.Tensor,
        output: torch.Tensor,
    ):
        self.iter += 1
        if self.iter % self.every_n != 0:
            return
        vals = output[0].detach().flatten()
        mean = vals.mean().cpu().item()
        std = vals.std().cpu().item()
        abs_perc90 = vals.abs().quantile(0.9).cpu().item()
        _max = vals.abs().max().cpu().item()
        frac_dead = (vals.abs() < self.threshold_dead).sum().cpu().item() / len(
            vals
        )

        self.hook.stats[self.name].append(
            GradientStats(mean, std, abs_perc90, _max, frac_dead)
        )


class ModelTelemetry(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        loss_train_every_n: int = 1,
        loss_test_every_n: int = 1,
        parameters_every_n: int = 1,
        activations_every_n: int = 1,
        gradients_every_n: int = 1,
        loss_names: T.Tuple[str] = ("loss",),
        activations_name_patterns: T.Tuple[str] = None,
        gradients_name_patterns: T.Tuple[str] = None,
        parameters_name_patterns: T.Tuple[str] = None,
        max_depth_search: int = 3,
    ):
        super().__init__()
        self.model = model

        # activations bit
        self.activations_history = ActivationsHistory(
            self.model,
            every_n=activations_every_n,
            name_patterns=activations_name_patterns,
            max_depth_search=max_depth_search,
        )

        # loss bit
        self.loss_history_train = LossHistory(
            loss_train_every_n, names=loss_names
        )
        self.loss_history_test = LossHistory(
            loss_test_every_n, names=loss_names
        )

        # parameter bit
        self.parameter_history = ParametersHistory(
            self.model,
            every_n=parameters_every_n,
            name_patterns=parameters_name_patterns,
            max_depth_search=max_depth_search,
        )

        # gradient bit
        self.gradients_history = GradientsHistory(
            self.model,
            every_n=gradients_every_n,
            name_patterns=gradients_name_patterns,
            max_depth_search=max_depth_search,
        )

    @property
    def name_matches_activations(self):
        return list(self.activations_history.name_matches)

    @property
    def name_matches_gradients(self):
        return list(self.gradients_history.name_matches)

    @property
    def name_matches_parameters(self):
        return list(self.parameter_history.name_matches)

    def forward(self, x: torch.Tensor):
        return self.model(x)

    def clean_hooks(self):
        self.activations_history.clean()
        self.gradients_history.clean()

    def draw_activation_stats(
        self,
        figsize: T.Tuple[int, int] = (12, 8),
        yscale: str = "linear",
        leg_lw: float = 5.0,
    ):
        fig, axs = plt.subplots(figsize=figsize, nrows=3, sharex=True)
        plt.suptitle("Activation Stats")

        # activation mean
        ax = axs[1]
        for _name, _stats in self.activations_history.stats.items():
            ax.plot([s.mean for s in _stats], label=_name, alpha=0.5)
        ax.set(title="mean", yscale=yscale)

        # activation std
        ax = axs[2]
        for _name, _stats in self.activations_history.stats.items():
            ax.plot([s.std for s in _stats], label=_name, alpha=0.5)
        ax.set(title="standard deviation", yscale=yscale)

        # share of dead neurons
        ax = axs[0]
        for _name, _stats in self.activations_history.stats.items():
            ax.plot([s.frac_dead for s in _stats], label=_name, alpha=0.5)
        ax.set(title="fraction of dead neurons", xlabel="iter")

        axs[1].legend()
        for leg_obj in axs[1].legend().legendHandles:
            leg_obj.set_linewidth(leg_lw)

        plt.tight_layout()

    def draw_gradient_stats(
        self,
        figsize: T.Tuple[int, int] = (12, 15),
        yscale: str = "linear",
        leg_lw: float = 5.0,
    ):
        fig, axs = plt.subplots(figsize=figsize, nrows=5, sharex=True)
        plt.suptitle("Gradient Stats")

        # gradient mean
        ax = axs[4]
        for _name, _stats in self.gradients_history.stats.items():
            ax.plot([s.mean for s in _stats], label=_name, alpha=0.5)
        ax.set(title="mean", yscale=yscale)

        # gradient std
        ax = axs[3]
        for _name, _stats in self.gradients_history.stats.items():
            ax.plot([s.std for s in _stats], label=_name, alpha=0.5)
        ax.set(title="standard deviation", yscale=yscale)

        # abs_perc90
        ax = axs[2]
        for _name, _stats in self.gradients_history.stats.items():
            ax.plot([s.abs_perc90 for s in _stats], label=_name, alpha=0.5)
        ax.legend()
        ax.set(title="90%(abs)", yscale=yscale)

        # vanishing
        ax = axs[1]
        for _name, _stats in self.gradients_history.stats.items():
            ax.plot([s.frac_dead for s in _stats], label=_name, alpha=0.5)
        ax.set(title="frac(dead)")

        # exploding
        ax = axs[0]
        for _name, _stats in self.gradients_history.stats.items():
            ax.plot([s.max for s in _stats], label=_name, alpha=0.5)
        ax.set(title="max(abs)", xlabel="iter", yscale=yscale)

        for leg_obj in axs[2].legend().legendHandles:
            leg_obj.set_linewidth(leg_lw)

        plt.tight_layout()

    def draw_loss_history_train(self, **kwargs):
        self.loss_history_train.draw("train", **kwargs)

    def draw_loss_history_test(self, **kwargs):
        if len(self.loss_history_test.history) == 0:
            logger.warning("No test loss history available")
            return
        self.loss_history_test.draw("test", **kwargs)

    def draw_parameter_stats(self, *names, **kwargs):
        if len(names) == 0:
            names = self.parameter_history.name_matches
        for name in names:
            try:
                self.parameter_history.draw(name, **kwargs)
            except HistoryException as e:
                logger.error(e)


class LossHistory:
    def __init__(self, every_n: int = 1, names: T.Tuple[str] = ("loss",)):
        self.names = names
        self.history = []
        self.iter = []
        self.every_n = every_n

    def __call__(
        self, losses: T.Union[torch.Tensor, T.Tuple[torch.Tensor]], _iter: int
    ):
        if _iter % self.every_n != 0:
            return
        if isinstance(losses, torch.Tensor):
            self.history.append(losses.item())
        else:
            self.history.append([l.item() for l in losses])
        self.iter.append(_iter)

    def get_df(self) -> pd.DataFrame:
        df = pd.DataFrame({"iter": self.iter})
        if len(self.names) == 1:
            df[self.names[0]] = self.history
        else:
            for i, name in enumerate(self.names):
                df[name] = [l[i] for l in self.history]

        return df

    def get_rolling_mean_df(self, window: int = 10) -> pd.DataFrame:
        df = self.get_df()
        df_roll = df.rolling(window=window, on="iter", min_periods=1).mean()
        if "iter" not in df_roll.columns:
            df_roll["iter"] = range(len(df_roll))
        return df_roll

    def draw(
        self,
        label: str,
        window: int = 10,
        figsize: T.Tuple[int, int] = (12, 4),
        yscale: str = "linear",
    ):
        df = self.get_df()
        df_roll = self.get_rolling_mean_df(window=window)

        for name in self.names:
            fig, ax = plt.subplots(figsize=figsize)
            sns.lineplot(
                data=df, x="iter", y=name, ax=ax, label=label, alpha=0.5
            )
            sns.lineplot(
                data=df_roll,
                x="iter",
                y=name,
                ax=ax,
                label=f"{label} (rolling mean)",
                alpha=0.5,
            )
            ax.set(
                xlabel="Iter",
                ylabel="Loss",
                title=f"Loss History: {name}",
                yscale=yscale,
            )

            plt.tight_layout()

        return fig, ax


@dataclass
class ActivationStats:
    mean: float
    std: float
    frac_dead: int


@dataclass
class ParameterStats:
    iter: int
    mean: float
    std: float
    abs_perc90: float


@dataclass
class GradientStats:
    mean: float
    std: float
    abs_perc90: float
    max: float
    frac_dead: float



# -*- coding: utf-8 -*-
# based on https://github.com/fastai/course22p2/blob/master/nbs/13_resnet.ipynb
import typing as T

import torch
import torch.nn as nn
import torchinfo
from einops.layers.torch import Rearrange

import random_neural_net_models.utils as utils

logger = utils.get_logger("resnet.py")


class ResBlock(nn.Module):
    def __init__(self, ni, nf, stride=1, ks=3):
        super().__init__()
        self.conv1 = nn.Conv2d(
            ni, nf, kernel_size=ks, stride=1, padding=ks // 2
        )
        self.bn1 = nn.BatchNorm2d(nf)
        self.act1 = nn.ReLU()

        self.conv2 = nn.Conv2d(
            nf, nf, kernel_size=ks, stride=stride, padding=ks // 2
        )
        self.bn2 = nn.BatchNorm2d(nf)

        self.convs = nn.Sequential(
            self.conv1,
            self.bn1,
            self.act1,
            self.conv2,
            self.bn2,
        )

        self.idconv = (
            nn.Identity()
            if ni == nf
            else nn.Conv2d(ni, nf, kernel_size=1, stride=1)
        )
        self.pool = (
            nn.Identity() if stride == 1 else nn.AvgPool2d(2, ceil_mode=True)
        )
        self.idconvs = nn.Sequential(self.idconv, self.pool)
        self.act_out = nn.ReLU()

    def forward(self, x: torch.Tensor):
        x_conv = self.convs(x)
        x_id = self.idconvs(x)
        return self.act_out(x_conv + x_id)


class ResNet(nn.Module):
    def __init__(
        self,
        nfs: T.Tuple[int] = (8, 16, 32, 64, 128, 256),
        ks: int = 3,
        n_classes: int = 10,
    ):
        super().__init__()

        layers = [ResBlock(1, nfs[0], stride=1, ks=ks)]
        for ni, nf in zip(nfs[:-1], nfs[1:]):
            layers.append(ResBlock(ni, nf, stride=2, ks=ks))

        self.resblocks = nn.Sequential(*layers)
        resblock_info = torchinfo.summary(self.resblocks, (1, 1, 28, 28))

        _, c, h, w = resblock_info.summary_list[0].output_size
        n_lin_in = c * h * w
        self.lin = nn.Linear(n_lin_in, n_classes)

        self.bn1 = nn.BatchNorm1d(n_classes)
        self.add_dim = Rearrange("b h w -> b 1 h w")
        self.flatten = Rearrange("b c h w -> b (c h w)")

        self.net = nn.Sequential(
            self.add_dim, self.resblocks, self.flatten, self.lin, self.bn1
        )

    def forward(self, x: torch.Tensor):
        return self.net(x)



# -*- coding: utf-8 -*-
import typing as T

import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from einops.layers.torch import Rearrange
from torch.utils.data import Dataset

import random_neural_net_models.utils as utils

logger = utils.get_logger("convolution_lecun1990.py")


class DigitsDataset(Dataset):
    def __init__(self, X: pd.DataFrame, y: pd.Series, edge: int = 28):
        self.X = X
        self.y = y
        self.edge = edge

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx: int) -> T.Tuple[torch.Tensor, int]:
        img = (
            torch.from_numpy(self.X.iloc[idx].values / 255.0)  # normalizing
            .reshape(self.edge, self.edge)
            .double()
        )
        label = int(self.y.iloc[idx])
        return (img, label)


def calc_conv_output_dim(input_dim, kernel_size, padding, stride):
    return int((input_dim - kernel_size + 2 * padding) / stride + 1)


def densify_y(y: torch.Tensor) -> torch.Tensor:
    new_y = F.one_hot(y, num_classes=10)
    new_y[new_y == 0] = -1
    return new_y.double()


class Tanh(nn.Module):
    def __init__(self, A: float = 1.716, S: float = 2 / 3):
        super().__init__()
        self.register_buffer("A", torch.tensor(A))
        self.register_buffer("S", torch.tensor(S))

    def forward(self, x: torch.Tensor):
        return self.A * torch.tanh(self.S * x)


class Conv2d(torch.nn.Module):
    def __init__(
        self,
        edge: int,
        n_in_channels: int = 1,
        n_out_channels: int = 1,
        kernel_width: int = 5,
        kernel_height: int = 5,
        stride: int = 1,
        padding: int = 0,
        dilation: int = 1,
        lecun_init: bool = True,
    ):
        super().__init__()

        self.register_buffer("edge", torch.tensor(edge))
        self.register_buffer("n_in_channels", torch.tensor(n_in_channels))
        self.register_buffer("n_out_channels", torch.tensor(n_out_channels))
        self.register_buffer("kernel_width", torch.tensor(kernel_width))
        self.register_buffer("kernel_height", torch.tensor(kernel_height))
        self.register_buffer("stride", torch.tensor(stride))
        self.register_buffer("padding", torch.tensor(padding))
        self.register_buffer("dilation", torch.tensor(dilation))

        self.weight = nn.Parameter(
            torch.empty(
                n_in_channels * kernel_width * kernel_height,
                n_out_channels,
                dtype=torch.double,
            )
        )
        self.bias = nn.Parameter(
            torch.empty(1, n_out_channels, 1, 1, dtype=torch.double)
        )

        # self.bias = rearrange(self.bias, "out_channels -> 1 out_channels 1 1")

        if lecun_init:
            s = 2.4 / (n_in_channels * kernel_width * kernel_height)
            self.weight.data.uniform_(-s, s)
            self.bias.data.uniform_(-s, s)

        else:
            self.weight.data.normal_(0, 1.0)
            self.bias.data.normal_(0, 1.0)

        self.unfold = torch.nn.Unfold(
            kernel_size=(kernel_height, kernel_width),
            dilation=dilation,
            padding=padding,
            stride=stride,
        )
        out_h = out_w = calc_conv_output_dim(
            edge, kernel_width, padding, stride
        )
        self.fold = torch.nn.Fold(
            output_size=(out_h, out_w),
            kernel_size=(1, 1),
            dilation=dilation,
            padding=0,
            stride=1,
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # inspired by: https://discuss.pytorch.org/t/make-custom-conv2d-layer-efficient-wrt-speed-and-memory/70175/2

        # (N,C,in_h,in_w) -> (N, C*kh*kw, num_patches)
        # N = batch_size, C = in_channels, kh = kernel_height, kw = kernel_width
        input_unfolded = self.unfold(input)

        input_unfolded = rearrange(
            input_unfolded, "N r num_patches -> N num_patches r"
        )

        output_unfolded = input_unfolded @ self.weight
        output_unfolded = rearrange(
            output_unfolded,
            "N num_patches out_channels -> N out_channels num_patches",
        )

        output = self.fold(output_unfolded)  # (N, out_channels, out_h, out_w)
        if self.bias is not None:
            output += self.bias

        return output


class Model(nn.Module):
    # based on LeCun et al. 1990, _Handwritten Digit Recognition: Applications of Neural Net Chips and Automatic Learning_, Neurocomputing, https://link.springer.com/chapter/10.1007/978-3-642-76153-9_35
    # inspired by https://einops.rocks/pytorch-examples.html
    def __init__(
        self,
        edge: int = 28,
        n_classes: int = 10,
        lecun_init: bool = True,
        lecun_act: bool = True,
        A: float = 1.716,
        S: float = 2 / 3,
    ):
        super().__init__()

        self.conv1 = Conv2d(
            edge=edge,
            n_in_channels=1,
            n_out_channels=12,
            kernel_width=5,
            kernel_height=5,
            stride=2,
            padding=2,
            lecun_init=lecun_init,
        )
        edge = edge // 2  # effect of stride

        self.conv2 = Conv2d(
            edge=edge,
            n_in_channels=12,
            n_out_channels=12,
            kernel_width=5,
            kernel_height=5,
            stride=2,
            padding=2,
            lecun_init=lecun_init,
        )
        edge = edge // 2  # effect of stride
        self.lin1 = nn.Linear(edge * edge * 12, 30)
        self.lin2 = nn.Linear(30, n_classes)

        if lecun_init:
            s = 2.4 / self.lin1.weight.shape[0]
            self.lin1.weight.data.uniform_(-s, s)

            s = 2.4 / self.lin2.weight.shape[0]
            self.lin2.weight.data.uniform_(-s, s)

        if lecun_act:
            self.act_conv1 = Tanh(A, S)
            self.act_conv2 = Tanh(A, S)
            self.act_lin1 = Tanh(A, S)
            self.act_lin2 = Tanh(A, S)
        else:
            self.act_conv1 = nn.Tanh()
            self.act_conv2 = nn.Tanh()
            self.act_lin1 = nn.Tanh()
            self.act_lin2 = nn.Tanh()

        self.net = nn.Sequential(
            Rearrange("b h w -> b 1 h w"),
            self.conv1,
            self.act_conv1,
            self.conv2,
            self.act_conv2,
            Rearrange("b c h w -> b (c h w)"),
            self.lin1,
            self.act_lin1,
            self.lin2,
            self.act_lin2,
        )

    def forward(self, x: torch.Tensor):
        return self.net(x)
